@@ -7,24 +7,41 @@ import logging
 import queue
 import threading
 import time
-from abc import abstractmethod
 from datetime import datetime
-from enum import Enum
 from functools import lru_cache
+from typing import Type
 
-from yarl import URL
+from aleph_message.models import (
+    ForgetContent,
+    MessageType,
+    AggregateContent,
+    PostContent,
+    StoreContent,
+    PostMessage,
+    Message,
+    ForgetMessage,
+    AlephMessage,
+    AggregateMessage,
+    StoreMessage,
+    ProgramMessage,
+)
+from pydantic import ValidationError
+
+from aleph_client.types import Account, StorageEnum, GenericMessage
+from .exceptions import MessageNotFoundError, MultipleMessagesError
+from .models import MessagesResponse
+from .utils import get_message_type_value
 
 logger = logging.getLogger(__name__)
 
 try:
     import magic  # type:ignore
 except ImportError:
-    logger.warning("Could not import library 'magic'")
+    logger.info("Could not import library 'magic', MIME type detection disabled")
     magic = None  # type:ignore
 
 from .conf import settings
 from typing import Optional, Iterable, Union, Any, Dict, List, AsyncIterable
-from typing_extensions import Protocol  # Python < 3.8
 
 import aiohttp
 from aiohttp import ClientSession
@@ -32,34 +49,6 @@ from aiohttp import ClientSession
 from aleph_message.models.program import ProgramContent, Encoding  # type: ignore
 
 
-class StorageEnum(str, Enum):
-    ipfs = "ipfs"
-    storage = "storage"
-
-
-# Use a protocol to avoid importing crypto libraries
-class Account(Protocol):
-    CHAIN: str
-    CURVE: str
-    private_key: Union[str, bytes]
-
-    @abstractmethod
-    async def sign_message(self, message: Dict) -> Dict:
-        ...
-
-    @abstractmethod
-    def get_address(self) -> str:
-        ...
-
-    @abstractmethod
-    def get_public_key(self) -> str:
-        ...
-
-    @abstractmethod
-    async def decrypt(self, content) -> bytes:
-        ...
-
-
 @lru_cache()
 def _get_fallback_session(thread_id: Optional[int]) -> ClientSession:
     if settings.API_UNIX_SOCKET:
@@ -74,42 +63,36 @@ def get_fallback_session() -> ClientSession:
     return _get_fallback_session(thread_id=thread_id)
 
 
-def wrap_async(func):
-    def func_caller(*args, **kwargs):
-        loop = asyncio.get_event_loop()
-        return loop.run_until_complete(func(*args, **kwargs))
-
-    return func_caller
-
-
 async def ipfs_push(
-    content, session: Optional[ClientSession] = None, api_server: str = settings.API_HOST
+    content,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
 ) -> str:
     session = session or get_fallback_session()
 
-    async with session.post(f"{api_server}/api/v0/ipfs/add_json", json=content) as resp:
+    url = f"{api_server}/api/v0/ipfs/add_json"
+    logger.debug(f"Pushing to IPFS on {url}")
+
+    async with session.post(url, json=content) as resp:
         resp.raise_for_status()
         return (await resp.json()).get("hash")
 
 
-sync_ipfs_push = wrap_async(ipfs_push)
-
-
 async def storage_push(
-    content, session: Optional[ClientSession] = None, api_server: str = settings.API_HOST
+    content,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
 ) -> str:
     session = session or get_fallback_session()
 
-    async with session.post(
-        f"{api_server}/api/v0/storage/add_json", json=content
-    ) as resp:
+    url = f"{api_server}/api/v0/storage/add_json"
+    logger.debug(f"Pushing to storage on {url}")
+
+    async with session.post(url, json=content) as resp:
         resp.raise_for_status()
         return (await resp.json()).get("hash")
 
 
-sync_storage_push = wrap_async(storage_push)
-
-
 async def ipfs_push_file(
     file_content,
     session: Optional[ClientSession] = None,
@@ -120,14 +103,14 @@ async def ipfs_push_file(
     data = aiohttp.FormData()
     data.add_field("file", file_content)
 
-    async with session.post(f"{api_server}/api/v0/ipfs/add_file", data=data) as resp:
+    url = f"{api_server}/api/v0/ipfs/add_file"
+    logger.debug(f"Pushing file to IPFS on {url}")
+
+    async with session.post(url, data=data) as resp:
         resp.raise_for_status()
         return (await resp.json()).get("hash")
 
 
-sync_ipfs_push_file = wrap_async(ipfs_push_file)
-
-
 async def storage_push_file(
     file_content,
     session: Optional[ClientSession] = None,
@@ -138,64 +121,79 @@ async def storage_push_file(
     data = aiohttp.FormData()
     data.add_field("file", file_content)
 
-    async with session.post(f"{api_server}/api/v0/storage/add_file", data=data) as resp:
+    url = f"{api_server}/api/v0/storage/add_file"
+    logger.debug(f"Posting file on {url}")
+
+    async with session.post(url, data=data) as resp:
         resp.raise_for_status()
         return (await resp.json()).get("hash")
 
 
-sync_storage_push_file = wrap_async(storage_push_file)
-
-
 async def broadcast(
-    message, session: Optional[ClientSession] = None, api_server: str = settings.API_HOST
-):
+    message,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
+) -> None:
+    """Broadcast a message on the Aleph network via pubsub for nodes to pick it up."""
     session = session or get_fallback_session()
 
+    url = f"{api_server}/api/v0/ipfs/pubsub/pub"
+    logger.debug(f"Posting message on {url}")
+
     async with session.post(
-        f"{api_server}/api/v0/ipfs/pubsub/pub",
+        url,
         json={"topic": "ALEPH-TEST", "data": json.dumps(message)},
     ) as response:
         response.raise_for_status()
         result = await response.json()
-        if result["status"] == "warning":
-            if 'failed' in result:
+        status = result.get("status")
+        if status == "success":
+            return
+        elif status == "warning":
+            if result.get("failed"):
                 # Requires recent version of Pyaleph
-                logger.warning(f"Message failed to publish on {result.get('failed')}")
+                if result["failed"] == ["p2p"]:
+                    logger.info(
+                        f"Message published on IPFS but failed to publish on P2P"
+                    )
+                else:
+                    logger.warning(
+                        f"Message published but failed on {result.get('failed')}"
+                    )
             else:
                 logger.warning(f"Message failed to publish on IPFS and/or P2P")
-        return (result).get("value")
-
-
-sync_broadcast = wrap_async(broadcast)
+        else:
+            raise ValueError(
+                f"Invalid response from server, status in missing or unknown: '{status}'"
+            )
 
 
 async def create_post(
     account: Account,
     post_content,
     post_type: str,
-    ref=None,
+    ref: Optional[str] = None,
     address: Optional[str] = settings.ADDRESS_TO_USE,
-    channel: str = "TEST",
+    channel: str = settings.DEFAULT_CHANNEL,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
     inline: bool = True,
-    storage_engine: str = "storage",
-):
+    storage_engine: StorageEnum = StorageEnum.storage,
+) -> PostMessage:
     address = address or account.get_address()
 
-    post = {
-        "type": post_type,
-        "address": address,
-        "content": post_content,
-        "time": time.time(),
-    }
-    if ref is not None:
-        post["ref"] = ref
+    content = PostContent(
+        type=post_type,
+        address=address,
+        content=post_content,
+        time=time.time(),
+        ref=ref,
+    )
 
     return await submit(
-        account,
-        post,
-        "POST",
+        account=account,
+        content=content.dict(exclude_none=True),
+        message_type=MessageType.post,
         channel=channel,
         api_server=api_server,
         session=session,
@@ -204,34 +202,36 @@ async def create_post(
     )
 
 
-sync_create_post = wrap_async(create_post)
-
-
 async def create_aggregate(
     account: Account,
     key,
     content,
     address: Optional[str] = settings.ADDRESS_TO_USE,
-    channel: str = "TEST",
+    channel: str = settings.DEFAULT_CHANNEL,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
-):
+    inline: bool = True,
+) -> AggregateMessage:
     address = address or account.get_address()
 
-    post = {"key": key, "address": address, "content": content, "time": time.time()}
+    content_ = AggregateContent(
+        key=key,
+        address=address,
+        content=content,
+        time=time.time(),
+    )
+
     return await submit(
-        account,
-        post,
-        "AGGREGATE",
+        account=account,
+        content=content_.dict(exclude_none=True),
+        message_type=MessageType.aggregate,
         channel=channel,
         api_server=api_server,
         session=session,
+        inline=inline,
     )
 
 
-sync_create_aggregate = wrap_async(create_aggregate)
-
-
 async def create_store(
     account: Account,
     address=settings.ADDRESS_TO_USE,
@@ -239,12 +239,12 @@ async def create_store(
     file_hash: Optional[str] = None,
     guess_mime_type: bool = False,
     ref: Optional[str] = None,
-    storage_engine="storage",
+    storage_engine=StorageEnum.storage,
     extra_fields: Optional[dict] = None,
-    channel: str = "TEST",
+    channel: str = settings.DEFAULT_CHANNEL,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
-):
+) -> StoreMessage:
     address = address or account.get_address()
     extra_fields = extra_fields or {}
 
@@ -252,38 +252,42 @@ async def create_store(
         if file_content is None:
             raise ValueError("Please specify at least a file_content or a file_hash")
 
-        if storage_engine == "storage":
+        if storage_engine == StorageEnum.storage:
             file_hash = await storage_push_file(
                 file_content, session=session, api_server=api_server
             )
-        elif storage_engine == "ipfs":
+        elif storage_engine == StorageEnum.ipfs:
             file_hash = await ipfs_push_file(
                 file_content, session=session, api_server=api_server
             )
         else:
             raise ValueError(f"Unknown storage engine: '{storage_engine}'")
 
+    assert file_hash, "File hash should be empty"
+
     if magic is None:
         pass
-    elif guess_mime_type is True and "mime_type" not in extra_fields:
+    elif file_content and guess_mime_type and ("mime_type" not in extra_fields):
         extra_fields["mime_type"] = magic.from_buffer(file_content, mime=True)
 
     if ref:
         extra_fields["ref"] = ref
 
-    store_content = {
+    values = {
         "address": address,
         "item_type": storage_engine,
         "item_hash": file_hash,
         "time": time.time(),
     }
     if extra_fields is not None:
-        store_content.update(extra_fields)
+        values.update(extra_fields)
+
+    content = StoreContent(**values)
 
     return await submit(
-        account,
-        store_content,
-        "STORE",
+        account=account,
+        content=content.dict(exclude_none=True),
+        message_type=MessageType.store,
         channel=channel,
         api_server=api_server,
         session=session,
@@ -291,58 +295,70 @@ async def create_store(
     )
 
 
-sync_create_store = wrap_async(create_store)
-
-
 async def create_program(
-        account: Account,
-        program_ref: str,
-        entrypoint: str,
-        runtime: str,
-        storage_engine: StorageEnum = StorageEnum.storage,
-        channel: str = "TEST",
-        address: Optional[str] = settings.ADDRESS_TO_USE,
-        session: Optional[ClientSession] = None,
-        api_server: str = settings.API_HOST,
-        memory: int = settings.DEFAULT_VM_MEMORY,
-        encoding: Encoding = Encoding.zip,
-        volumes: List[Dict] = None,
-):
+    account: Account,
+    program_ref: str,
+    entrypoint: str,
+    runtime: str,
+    environment_variables: Optional[Dict[str, str]] = None,
+    storage_engine: StorageEnum = StorageEnum.storage,
+    channel: str = settings.DEFAULT_CHANNEL,
+    address: Optional[str] = settings.ADDRESS_TO_USE,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
+    memory: int = settings.DEFAULT_VM_MEMORY,
+    vcpus: int = settings.DEFAULT_VM_VCPUS,
+    timeout_seconds: float = settings.DEFAULT_VM_TIMEOUT,
+    persistent: bool = False,
+    encoding: Encoding = Encoding.zip,
+    volumes: List[Dict] = None,
+    subscriptions: Optional[List[Dict]] = None,
+) -> ProgramMessage:
     volumes = volumes if volumes is not None else []
 
     address = address or account.get_address()
 
     # TODO: Check that program_ref, runtime and data_ref exist
 
-    content = ProgramContent(**{
-        "type": "vm-function",
-        "address": address,
-        "allow_amend": False,
-        "code": {
-            "encoding": encoding,
-            "entrypoint": entrypoint,
-            "ref": program_ref,
-            "use_latest": True,
-        },
-        "on": {
-            "http": True,
-        },
-        "environment": {
-            "reproducible": False,
-            "internet": True,
-            "aleph_api": True,
-        },
-        "resources": {
-            "vcpus": 1,
-            "memory": memory,
-            "seconds": 30,
-        },
-        "runtime": {
-            "ref": runtime,
-            "use_latest": True,
-            "comment": "Aleph Alpine Linux with Python 3.8",
-        },
-        "volumes": volumes,
+    ## Register the different ways to trigger a VM
+    if subscriptions:
+        # Trigger on HTTP calls and on Aleph message subscriptions.
+        triggers = {"http": True, "persistent": persistent, "message": subscriptions}
+    else:
+        # Trigger on HTTP calls.
+        triggers = {"http": True, "persistent": persistent}
+
+    content = ProgramContent(
+        **{
+            "type": "vm-function",
+            "address": address,
+            "allow_amend": False,
+            "code": {
+                "encoding": encoding,
+                "entrypoint": entrypoint,
+                "ref": program_ref,
+                "use_latest": True,
+            },
+            "on": triggers,
+            "environment": {
+                "reproducible": False,
+                "internet": True,
+                "aleph_api": True,
+            },
+            "variables": environment_variables,
+            "resources": {
+                "vcpus": vcpus,
+                "memory": memory,
+                "seconds": timeout_seconds,
+            },
+            "runtime": {
+                "ref": runtime,
+                "use_latest": True,
+                "comment": "Official Aleph runtime"
+                if runtime == settings.DEFAULT_RUNTIME_ID
+                else "",
+            },
+            "volumes": volumes,
             # {
             #     "mount": "/opt/venv",
             #     "ref": "5f31b0706f59404fad3d0bff97ef89ddf24da4761608ea0646329362c662ba51",
@@ -355,13 +371,47 @@ async def create_program(
             #     "persistence": "host",
             #     "size_mib": 5
             # }
-        "time": time.time(),
-    })
+            "time": time.time(),
+        }
+    )
+
+    # Ensure that the version of aleph-message used supports the field.
+    assert content.on.persistent == persistent
 
     return await submit(
         account=account,
         content=content.dict(exclude_none=True),
-        message_type="PROGRAM",
+        message_type=MessageType.program,
+        channel=channel,
+        api_server=api_server,
+        storage_engine=storage_engine,
+        session=session,
+    )
+
+
+async def forget(
+    account: Account,
+    hashes: List[str],
+    reason: Optional[str],
+    storage_engine: StorageEnum = StorageEnum.storage,
+    channel: str = settings.DEFAULT_CHANNEL,
+    address: Optional[str] = settings.ADDRESS_TO_USE,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
+) -> ForgetMessage:
+    address = address or account.get_address()
+
+    content = ForgetContent(
+        hashes=hashes,
+        reason=reason,
+        address=address,
+        time=time.time(),
+    )
+
+    return await submit(
+        account,
+        content=content.dict(exclude_none=True),
+        message_type=MessageType.forget,
         channel=channel,
         api_server=api_server,
         storage_engine=storage_engine,
@@ -370,46 +420,16 @@ async def create_program(
     )
 
 
-def sync_create_program(
-        account: Account,
-        program_ref: str,
-        entrypoint: str,
-        runtime: str,
-        storage_engine: StorageEnum = StorageEnum.storage,
-        channel: str = "TEST",
-        address: Optional[str] = settings.ADDRESS_TO_USE,
-        session: Optional[ClientSession] = None,
-        api_server: str = settings.API_HOST,
-        memory: int = settings.DEFAULT_VM_MEMORY,
-        encoding: Encoding = Encoding.zip,
-        volumes: List[Dict] = None,
-):
-    return wrap_async(create_program)(
-        account=account,
-        program_ref=program_ref,
-        entrypoint=entrypoint,
-        runtime=runtime,
-        storage_engine=storage_engine,
-        channel=channel,
-        address=address,
-        session=session,
-        api_server=api_server,
-        memory=memory,
-        encoding=encoding,
-        volumes=volumes,
-    )
-
-
 async def submit(
     account: Account,
-    content: dict,
-    message_type: str,
-    channel: str = "IOT_TEST",
+    content: Dict,
+    message_type: MessageType,
+    channel: str = settings.DEFAULT_CHANNEL,
     api_server: str = settings.API_HOST,
-    storage_engine: str = "storage",
+    storage_engine: StorageEnum = StorageEnum.storage,
     session: Optional[ClientSession] = None,
     inline: bool = True,
-):
+) -> AlephMessage:
     message: Dict[str, Any] = {
         #'item_hash': ipfs_hash,
         "chain": account.CHAIN,
@@ -421,63 +441,76 @@ async def submit(
 
     item_content: str = json.dumps(content, separators=(",", ":"))
 
-    if inline and (len(item_content) < 100000):
+    if inline and (len(item_content) < 50000):
         message["item_content"] = item_content
         h = hashlib.sha256()
         h.update(message["item_content"].encode("utf-8"))
         message["item_hash"] = h.hexdigest()
+        message["item_type"] = "inline"
     else:
-        if storage_engine == "ipfs":
-            message["item_hash"] = await ipfs_push(content, api_server=api_server)
+        if storage_engine == StorageEnum.ipfs:
+            message["item_hash"] = await ipfs_push(
+                content, session=session, api_server=api_server
+            )
         else:  # storage
-            message["item_hash"] = await storage_push(content, api_server=api_server)
+            assert storage_engine == StorageEnum.storage
+            message["item_hash"] = await storage_push(
+                content, session=session, api_server=api_server
+            )
 
     message = await account.sign_message(message)
     await broadcast(message, session=session, api_server=api_server)
 
     # let's add the content to the object so users can access it.
     message["content"] = content
-    return message
 
-
-sync_submit = wrap_async(submit)
+    return Message(**message)
 
 
 async def fetch_aggregate(
     address: str,
-    key,
+    key: str,
+    limit: Optional[int] = 100,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
-):
+) -> Dict[str, Dict]:
     session = session or get_fallback_session()
 
+    params: Dict[str, Any] = {"keys": key}
+    if limit:
+        params["limit"] = limit
+
     async with session.get(
-        f"{api_server}/api/v0/aggregates/{address}.json?keys={key}"
+        f"{api_server}/api/v0/aggregates/{address}.json", params=params
     ) as resp:
-        return (await resp.json()).get("data", dict()).get(key)
-
-
-sync_fetch_aggregate = wrap_async(fetch_aggregate)
+        result = await resp.json()
+        data = result.get("data", dict())
+        return data.get(key)
 
 
 async def fetch_aggregates(
     address: str,
     keys: Optional[Iterable[str]] = None,
+    limit: Optional[int] = 100,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
 ) -> Dict[str, Dict]:
     session = session or get_fallback_session()
 
     keys_str = ",".join(keys) if keys else ""
-    query_string = f"?keys={keys_str}" if keys else ""
+    params: Dict[str, Any] = {}
+    if keys_str:
+        params["keys"] = keys_str
+    if limit:
+        params["limit"] = limit
 
     async with session.get(
-        f"{api_server}/api/v0/aggregates/{address}.json{query_string}"
+        f"{api_server}/api/v0/aggregates/{address}.json",
+        params=params,
     ) as resp:
-        return (await resp.json()).get("data", dict())
-
-
-sync_fetch_aggregates = wrap_async(fetch_aggregates)
+        result = await resp.json()
+        data = result.get("data", dict())
+        return data
 
 
 async def get_posts(
@@ -489,6 +522,7 @@ async def get_posts(
     tags: Optional[Iterable[str]] = None,
     hashes: Optional[Iterable[str]] = None,
     channels: Optional[Iterable[str]] = None,
+    chains: Optional[Iterable[str]] = None,
     start_date: Optional[Union[datetime, float]] = None,
     end_date: Optional[Union[datetime, float]] = None,
     session: Optional[ClientSession] = None,
@@ -510,47 +544,52 @@ async def get_posts(
         params["hashes"] = ",".join(hashes)
     if channels is not None:
         params["channels"] = ",".join(channels)
+    if chains is not None:
+        params["chains"] = ",".join(chains)
 
     if start_date is not None:
         if not isinstance(start_date, float) and hasattr(start_date, "timestamp"):
             start_date = start_date.timestamp()
-        params["start_date"] = start_date
+        params["startDate"] = start_date
     if end_date is not None:
         if not isinstance(end_date, float) and hasattr(start_date, "timestamp"):
             end_date = end_date.timestamp()
-        params["end_date"] = end_date
+        params["endDate"] = end_date
 
     async with session.get(f"{api_server}/api/v0/posts.json", params=params) as resp:
         resp.raise_for_status()
         return await resp.json()
 
 
-sync_get_posts = wrap_async(get_posts)
-
-
 async def get_messages(
     pagination: int = 200,
     page: int = 1,
-    message_type: Optional[str] = None,
+    message_type: Optional[MessageType] = None,
     content_types: Optional[Iterable[str]] = None,
+    content_keys: Optional[Iterable[str]] = None,
     refs: Optional[Iterable[str]] = None,
     addresses: Optional[Iterable[str]] = None,
     tags: Optional[Iterable[str]] = None,
     hashes: Optional[Iterable[str]] = None,
     channels: Optional[Iterable[str]] = None,
+    chains: Optional[Iterable[str]] = None,
     start_date: Optional[Union[datetime, float]] = None,
     end_date: Optional[Union[datetime, float]] = None,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
-) -> Dict[str, Any]:
+    ignore_invalid_messages: bool = True,
+    invalid_messages_log_level: int = logging.NOTSET,
+) -> MessagesResponse:
     session = session or get_fallback_session()
 
     params: Dict[str, Any] = dict(pagination=pagination, page=page)
 
     if message_type is not None:
-        params["msgType"] = message_type
+        params["msgType"] = message_type.value
     if content_types is not None:
         params["contentTypes"] = ",".join(content_types)
+    if content_keys is not None:
+        params["contentKeys"] = ",".join(content_keys)
     if refs is not None:
         params["refs"] = ",".join(refs)
     if addresses is not None:
@@ -561,49 +600,107 @@ async def get_messages(
         params["hashes"] = ",".join(hashes)
     if channels is not None:
         params["channels"] = ",".join(channels)
+    if chains is not None:
+        params["chains"] = ",".join(chains)
 
     if start_date is not None:
         if not isinstance(start_date, float) and hasattr(start_date, "timestamp"):
             start_date = start_date.timestamp()
-        params["start_date"] = start_date
+        params["startDate"] = start_date
     if end_date is not None:
         if not isinstance(end_date, float) and hasattr(start_date, "timestamp"):
             end_date = end_date.timestamp()
-        params["end_date"] = end_date
+        params["endDate"] = end_date
 
     async with session.get(f"{api_server}/api/v0/messages.json", params=params) as resp:
         resp.raise_for_status()
-        return await resp.json()
+        response_json = await resp.json()
+        messages_raw = response_json["messages"]
+
+        # All messages may not be valid according to the latest specification in
+        # aleph-message. This allows the user to specify how errors should be handled.
+        messages: List[AlephMessage] = []
+        for message_raw in messages_raw:
+            try:
+                message = Message(**message_raw)
+                messages.append(message)
+            except KeyError as e:
+                if not ignore_invalid_messages:
+                    raise e
+                logger.log(
+                    level=invalid_messages_log_level,
+                    msg=f"KeyError: Field '{e.args[0]}' not found",
+                )
+            except ValidationError as e:
+                if not ignore_invalid_messages:
+                    raise e
+                if invalid_messages_log_level:
+                    logger.log(level=invalid_messages_log_level, msg=e)
+
+        return MessagesResponse(
+            messages=messages,
+            pagination_page=response_json["pagination_page"],
+            pagination_total=response_json["pagination_total"],
+            pagination_per_page=response_json["pagination_per_page"],
+            pagination_item=response_json["pagination_item"],
+        )
 
 
-sync_get_messages = wrap_async(get_messages)
+async def get_message(
+    item_hash: str,
+    message_type: Optional[Type[GenericMessage]] = None,
+    channel: Optional[str] = None,
+    session: Optional[ClientSession] = None,
+    api_server: str = settings.API_HOST,
+) -> GenericMessage:
+    """Get a single message from its `item_hash`."""
+    messages_response = await get_messages(
+        hashes=[item_hash],
+        session=session,
+        channels=[channel] if channel else None,
+        api_server=api_server,
+    )
+    if len(messages_response.messages) < 1:
+        raise MessageNotFoundError(f"No such hash {item_hash}")
+    if len(messages_response.messages) != 1:
+        raise MultipleMessagesError(
+            f"Multiple messages found for the same item_hash `{item_hash}`"
+        )
+    message: GenericMessage = messages_response.messages[0]
+    if message_type:
+        expected_type = get_message_type_value(message_type)
+        if message.type != expected_type:
+            raise TypeError(
+                f"The message type '{message.type}' "
+                f"does not match the expected type '{expected_type}'"
+            )
+    return message
 
 
 async def watch_messages(
-    pagination: int = 200,
-    page: int = 1,
-    message_type: Optional[str] = None,
+    message_type: Optional[MessageType] = None,
     content_types: Optional[Iterable[str]] = None,
     refs: Optional[Iterable[str]] = None,
     addresses: Optional[Iterable[str]] = None,
     tags: Optional[Iterable[str]] = None,
     hashes: Optional[Iterable[str]] = None,
     channels: Optional[Iterable[str]] = None,
+    chains: Optional[Iterable[str]] = None,
     start_date: Optional[Union[datetime, float]] = None,
     end_date: Optional[Union[datetime, float]] = None,
     session: Optional[ClientSession] = None,
     api_server: str = settings.API_HOST,
-) -> AsyncIterable[Dict[str, Any]]:
+) -> AsyncIterable[AlephMessage]:
     """
     Iterate over current and future matching messages asynchronously.
     """
 
     session = session or get_fallback_session()
 
-    params: Dict[str, Any] = dict(pagination=pagination, page=page)
+    params: Dict[str, Any] = dict()
 
     if message_type is not None:
-        params["msgType"] = message_type
+        params["msgType"] = message_type.value
     if content_types is not None:
         params["contentTypes"] = ",".join(content_types)
     if refs is not None:
@@ -616,32 +713,30 @@ async def watch_messages(
         params["hashes"] = ",".join(hashes)
     if channels is not None:
         params["channels"] = ",".join(channels)
+    if chains is not None:
+        params["chains"] = ",".join(chains)
 
     if start_date is not None:
         if not isinstance(start_date, float) and hasattr(start_date, "timestamp"):
             start_date = start_date.timestamp()
-        params["start_date"] = start_date
+        params["startDate"] = start_date
     if end_date is not None:
         if not isinstance(end_date, float) and hasattr(start_date, "timestamp"):
             end_date = end_date.timestamp()
-        params["end_date"] = end_date
+        params["endDate"] = end_date
 
-    # FIXME:
-    #  We build the URL manually since aiohttp.ClientSession.ws_connect does not support
-    #  the `params` argument at the moment.
-    #  Upstream issue: https://github.com/aio-libs/aiohttp/issues/5868
-    #  Upstream pull request: https://github.com/aio-libs/aiohttp/pull/5869
-    url = URL(f"{api_server}/api/ws0/messages").with_query(params)
-
-    async with session.ws_connect(url) as ws:
+    async with session.ws_connect(
+        f"{api_server}/api/ws0/messages", params=params
+    ) as ws:
         logger.debug("Websocket connected")
         async for msg in ws:
             if msg.type == aiohttp.WSMsgType.TEXT:
-                if msg.data == 'close cmd':
+                if msg.data == "close cmd":
                     await ws.close()
                     break
                 else:
-                    yield json.loads(msg.data)
+                    data = json.loads(msg.data)
+                    yield Message(**data)
             elif msg.type == aiohttp.WSMsgType.ERROR:
                 break
 
@@ -653,21 +748,7 @@ async def _run_watch_messages(coroutine: AsyncIterable, output_queue: queue.Queu
 
 
 def _start_run_watch_messages(output_queue: queue.Queue, args: List, kwargs: Dict):
-    """Thread entrypoint to run the `watch_messages` asynchronous generator in a thread.
-    """
+    """Thread entrypoint to run the `watch_messages` asynchronous generator in a thread."""
     watcher = watch_messages(*args, **kwargs)
     runner = _run_watch_messages(watcher, output_queue)
-    asyncio.new_event_loop().run_until_complete(runner)
-
-
-def sync_watch_messages(*args, **kwargs):
-    """
-    Iterate over current and future matching messages synchronously.
-
-    Runs the `watch_messages` asynchronous generator in a thread.
-    """
-    output_queue = queue.Queue()
-    thread = threading.Thread(target=_start_run_watch_messages, args=(output_queue, args, kwargs))
-    thread.start()
-    while True:
-        yield output_queue.get()
+    asyncio.run(runner)
