@@ -1,20 +1,16 @@
-import json
-import ecdsa
-import hashlib
 import base64
+import hashlib
+import json
+
+import ecdsa
+from cosmospy._wallet import privkey_to_address, privkey_to_pubkey
 
-from substrateinterface import Keypair
 from .common import (
     BaseAccount,
     get_fallback_private_key,
     get_verification_buffer,
-    get_public_key,
-    PRIVATE_KEY_FILE,
 )
 
-from cosmospy._wallet import privkey_to_address, privkey_to_pubkey
-from cosmospy.typing import SyncMode
-
 DEFAULT_HRP = "cosmos"
 
 
@@ -56,7 +52,9 @@ class CSDKAccount(BaseAccount):
         self.private_key = private_key
         self.hrp = hrp
 
-    def sign_message(self, message):
+    async def sign_message(self, message):
+        message = self._setup_sender(message)
+
         verif = get_verification_string(message)
 
         privkey = ecdsa.SigningKey.from_string(self.private_key, curve=ecdsa.SECP256k1)
@@ -66,7 +64,7 @@ class CSDKAccount(BaseAccount):
             sigencode=ecdsa.util.sigencode_string_canonize,
         )
         signature_base64_str = base64.b64encode(signature_compact).decode("utf-8")
-        base64_pubkey = base64.b64encode(self.get_public_key()).decode("utf-8")
+        base64_pubkey = base64.b64encode(self.get_public_key().encode()).decode("utf-8")
 
         sig = {
             "signature": signature_base64_str,
@@ -77,11 +75,11 @@ class CSDKAccount(BaseAccount):
         message["signature"] = json.dumps(sig)
         return message
 
-    def get_address(self):
+    def get_address(self) -> str:
         return privkey_to_address(self.private_key)
 
-    def get_public_key(self):
-        return privkey_to_pubkey(self.private_key)
+    def get_public_key(self) -> str:
+        return privkey_to_pubkey(self.private_key).decode()
 
 
 def get_fallback_account(hrp=DEFAULT_HRP):
