@@ -1,19 +1,20 @@
 """ Barebone NULS address and message signing support.
 """
-import struct
 import hashlib
-from secp256k1 import PrivateKey, PublicKey
+import logging
+import struct
 from binascii import hexlify, unhexlify
+from typing import Optional
+
+from coincurve import PrivateKey, PublicKey
+
 from .common import (
     get_public_key,
-    generate_key,
     get_verification_buffer,
     BaseAccount,
     get_fallback_private_key,
 )
 
-import logging
-
 LOGGER = logging.getLogger("NULS")
 
 PLACE_HOLDER = b"\xFF\xFF\xFF\xFF"
@@ -111,7 +112,7 @@ def read_by_length(buffer, cursor=0, check_size=True):
         size = 1
 
     value = buffer[cursor + size : cursor + size + length]
-    return (size + length, value)
+    return size + length, value
 
 
 def write_with_length(buffer):
@@ -225,14 +226,29 @@ class BaseNulsData:
 
 class NulsSignature(BaseNulsData):
     ALG_TYPE = 0  # only one for now...
+    pub_key: Optional[bytes]
+    digest_bytes: Optional[bytes]
+    sig_ser: Optional[bytes]
+    ecc_type: Optional[bytes]
 
     def __init__(self, data=None):
         self.pub_key = None
         self.digest_bytes = None
         self.sig_ser = None
+        self.ecc_type = None
         if data is not None:
             self.parse(data)
 
+    def __eq__(self, other):
+        return all(
+            (
+                (self.pub_key == other.pub_key),
+                (self.digest_bytes == other.digest_bytes),
+                (self.sig_ser == other.sig_ser),
+                (self.ecc_type == other.ecc_type),
+            )
+        )
+
     def parse(self, buffer, cursor=0):
         pos, self.pub_key = read_by_length(buffer, cursor)
         cursor += pos
@@ -243,27 +259,24 @@ class NulsSignature(BaseNulsData):
         return cursor
 
     @classmethod
-    def sign_data(cls, pri_key, digest_bytes):
-        privkey = PrivateKey(
-            pri_key, raw=True
-        )  # we expect to have a private key as bytes. unhexlify it before passing.
+    def sign_data(cls, pri_key: bytes, digest_bytes: bytes):
+        privkey = PrivateKey(pri_key)
+        # we expect to have a private key as bytes. unhexlify it before passing.
         item = cls()
-        item.pub_key = privkey.pubkey.serialize()
+        item.pub_key = privkey.public_key.format()
         item.digest_bytes = digest_bytes
-        sig_check = privkey.ecdsa_sign(digest_bytes, raw=True)
-        item.sig_ser = privkey.ecdsa_serialize(sig_check)
+        item.sig_ser = privkey.sign(digest_bytes, hasher=None)
         return item
 
     @classmethod
-    def sign_message(cls, pri_key, message):
+    async def sign_message(cls, pri_key: bytes, message):
         # we expect to have a private key as bytes. unhexlify it before passing
-        privkey = PrivateKey(pri_key, raw=True)
+        privkey = PrivateKey(pri_key)
         item = cls()
         message = VarInt(len(message)).encode() + message
-        item.pub_key = privkey.pubkey.serialize()
+        item.pub_key = privkey.public_key.format()
         # item.digest_bytes = digest_bytes
-        sig_check = privkey.ecdsa_sign(MESSAGE_TEMPLATE.format(message).encode())
-        item.sig_ser = privkey.ecdsa_serialize(sig_check)
+        item.sig_ser = privkey.sign(MESSAGE_TEMPLATE.format(message).encode())
         return item
 
     def serialize(self, with_length=False):
@@ -277,12 +290,11 @@ class NulsSignature(BaseNulsData):
             return output
 
     def verify(self, message):
-        pub = PublicKey(self.pub_key, raw=True)
+        pub = PublicKey(self.pub_key)
         message = VarInt(len(message)).encode() + message
         # LOGGER.debug("Comparing with %r" % (MESSAGE_TEMPLATE.format(message).encode()))
         try:
-            sig_raw = pub.ecdsa_deserialize(self.sig_ser)
-            good = pub.ecdsa_verify(MESSAGE_TEMPLATE.format(message).encode(), sig_raw)
+            good = pub.verify(self.sig_ser, MESSAGE_TEMPLATE.format(message).encode())
         except Exception:
             LOGGER.exception("Verification failed")
             good = False
@@ -335,10 +347,13 @@ class NULSAccount(BaseAccount):
         self.private_key = private_key
         self.chain_id = chain_id
 
-    def sign_message(self, message):
+    async def sign_message(self, message):
+        message = self._setup_sender(message)
+
         sig = NulsSignature.sign_message(
             self.private_key, get_verification_buffer(message)
         )
+
         message["signature"] = sig.serialize().hex()
         return message
 
