@@ -1,16 +1,14 @@
 import base64
+
 from nuls2.model.data import (
-    read_by_length,
-    write_with_length,
     public_key_to_hash,
     address_from_hash,
     sign_recoverable_message,
     NETWORKS,
-    recover_message_address,
 )
+
 from .common import (
     get_public_key,
-    generate_key,
     get_verification_buffer,
     BaseAccount,
     get_fallback_private_key,
@@ -38,9 +36,11 @@ class NULSAccount(BaseAccount):
         else:
             self.prefix = prefix
 
-    def sign_message(self, message):
+    async def sign_message(self, message):
         # sig = NulsSignature.sign_message(self.private_key,
         #                                  get_verification_buffer(message))
+        message = self._setup_sender(message)
+
         sig = sign_recoverable_message(
             self.private_key, get_verification_buffer(message)
         )
