@@ -1,13 +1,12 @@
 import json
 
 from substrateinterface import Keypair
+
 from .common import (
     BaseAccount,
-    get_fallback_private_key,
     get_verification_buffer,
-    get_public_key,
-    PRIVATE_KEY_FILE,
 )
+from ..conf import settings
 
 
 class DOTAccount(BaseAccount):
@@ -21,7 +20,8 @@ class DOTAccount(BaseAccount):
             self.mnemonics, address_type=address_type
         )
 
-    def sign_message(self, message):
+    async def sign_message(self, message):
+        message = self._setup_sender(message)
         verif = get_verification_buffer(message).decode("utf-8")
         sig = {"curve": self.CURVE, "data": self._account.sign(verif)}
         message["signature"] = json.dumps(sig)
@@ -40,11 +40,11 @@ def get_fallback_account():
 
 def get_fallback_mnemonics():
     try:
-        with open(PRIVATE_KEY_FILE, "r") as prvfile:
+        with open(settings.PRIVATE_KEY_FILE, "r") as prvfile:
             mnemonic = prvfile.read()
     except OSError:
         mnemonic = Keypair.generate_mnemonic()
-        with open(PRIVATE_KEY_FILE, "w") as prvfile:
+        with open(settings.PRIVATE_KEY_FILE, "w") as prvfile:
             prvfile.write(mnemonic)
 
     return mnemonic
